/* $Id: piim.cpp 61 2006-07-31 21:15:25Z pjohnson $
  ------------------------------------------------------------------------
  This file is part of the PIIM package; please see:
    Johnson, PLF and Slatkin, M. 2006. "Inference of population
    genetic parameters in metagenomics: a clean look at messy data."
    Genome Research (in press).
 
  Copyright 2006 The Regents of the University of California.  All
  rights reserved.
 
  IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT,
  INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
  LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
  DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE POSSIBILITY
  OF SUCH DAMAGE.
 
  REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING
  DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
  IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
  UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 
  This software is distributed under the terms of Version 2 of the
  GNU General Public License as published by the Free Software
  Foundation.  If you have not received a copy of Version 2 of the
  GNU General Public License with this program, you may download the
  GPL from the following url: http://www.gnu.org/copyleft/gpl.html.


  ---------------------------------------------------------------------------
  Inputs: one assembled contig containing the followed detailed info:
     1. vector of quality scores at each position
     2. boolean vector at each position indicating whether that
        chromosome has the apparent MINOR or apparent MAJOR allele

  ---------------------------------------------------------------------------
*/

#include "util.hpp"
#include "gsl_util.hpp"
#include "freqspec.hpp"
#include <string>
#include <vector>
#include <map>
#include <algorithm>
#include <fstream>
#include <sstream>
#include <iostream>
#include <stdexcept>

#include <assert.h>
#include <math.h>
#include <gsl/gsl_sf_gamma.h>

#include <libxml++/libxml++.h>

using namespace std;

/*---------------------------------------------------------------------------*/
// Generic mapping class for discrete probability distribution.
// (Perhaps I actually want to use a STL map?)
class CProbDist {
public:
    CProbDist(int rangeMin, int rangeMax) :
        m_RangeMin(rangeMin), m_RangeMax(rangeMax),
        m_ProbDist(rangeMax-rangeMin+1) {}
    double& SetProb(int i) { return m_ProbDist[i - m_RangeMin]; }
    const double& GetProb(int i) const { return m_ProbDist[i - m_RangeMin]; }

    void Serialize(ostream &o) const {
        for (TProbs::const_iterator i = m_ProbDist.begin();
             i != m_ProbDist.end();  ++i) { o << *i << " "; }
    }
    void DeSerialize(istream &in) {
        double p;
        for (unsigned int i = 0;  i < m_ProbDist.size();  ++i) {
            in >> p;
            if (!in.good()) {
                throw runtime_error("CProbDist serialization corrupted!");
            }
            m_ProbDist[i] = p;
        }
    }

private:
    typedef vector<double> TProbs;
    int m_RangeMin;
    int m_RangeMax;
    TProbs m_ProbDist;
};


//---------------------------------------------------------------------------
// Container classes to hold quality scores
class CQScoreColumn {
public:
    typedef vector<unsigned char> TScores[4];
    typedef vector<bool> TErrors[4];
    enum EWhichScores {
        eMajor, eMinor, eOther1, eOther2, eAll
    };

public:
    CQScoreColumn(int contigPos, const TScores &scores,
                  const string &prObservedCond, bool knowAncestral,
                  const TErrors *errors = NULL);
    void WriteToXML(xmlpp::Element *root, bool knowAncestral) const;

    double GetPrObservedCond(int m) const;
    double GetPrPrior(int m, const CFreqSpectrum &fs) const {
        return fs(m, GetDepth()); }
    double GetPrObserved(const CFreqSpectrum &) const;
    double GetPrPosterior(int m, const CFreqSpectrum &fs) const {
        return GetPrObservedCond(m) * GetPrPrior(m, fs) / GetPrObserved(fs);
    }

    int GetContigPos(void) const { return m_ContigPos; }

    unsigned int GetDepth(EWhichScores w = eAll) const {
        if (w == eAll) {
            int tot = 0;
            for (int i = 0;  i < 4;  ++i) {
                tot += m_Scores[i].size();
            }
            return tot;
        } else {
            return m_Scores[w].size();
        }
    }

    double GetErrProb(unsigned int i, EWhichScores w = eAll) const {
        double score;
        if (w == eAll) {
            int j;
            for (j = 0;  j < 4;  ++j) {
                if (i < m_Scores[j].size()) {
                    score = m_Scores[j][i];
                    break;
                } else {
                    i -= m_Scores[j].size();
                }
            }
            assert(j < 4);
        } else {
            assert(i < m_Scores[w].size());
            score = m_Scores[w][i];
        }
        return pow(10, -score/10.0);
    }
    bool GetErrTruth(unsigned int i, EWhichScores w) const {
        if (w == eAll) {
            throw runtime_error("eAll case given to GetErrTruth");
        } else {
            assert(i < m_Errors[w].size());
            return m_Errors[w][i];
        }
    }
protected:
    void x_CalcErrorDist(EWhichScores, CProbDist &errorPrbs) const;
    void x_CalcObservedCondDist(void);

private:
    typedef vector<bool> TState;
    static bool x_IncrementState (TState &state);

private:
    int     m_ContigPos;  //position in contig
    TScores m_Scores;
    TErrors m_Errors;//used for simulations, where we know truth

    CProbDist m_pObservedCond;
};

/*---------------------------------------------------------------------------*/

class CQScores {
public:
    CQScores(CFreqSpectrum *fs) : m_FS(fs), m_KnowAncestral(false) {}
    ~CQScores(void) {
        for (TColumns::iterator colI = m_Columns.begin();
             colI != m_Columns.end();  ++colI) {
            delete *colI;
        }
    }

    bool& SetKnowAncestral(void) { return m_KnowAncestral; }
    bool GetKnowAncestral(void) const { return m_KnowAncestral; }

    const CFreqSpectrum& GetFreqSpectrum(void) const { return *m_FS; }
    auto_ptr<CFreqSpectrum>& SetFreqSpectrum(void) { return m_FS; }

    void LoadFromStream(istream &in);
    void WriteToStream(ostream &out);
    void OptimizeParamsSimplex(void) { CFSMaximizer(*this); }
    void OptimizeParamsGrid(const CFreqSpectrum::TParams& start,
                            const CFreqSpectrum::TParams& stop,
                            const CFreqSpectrum::TParams& inc);

    unsigned int GetNumColumns(void) const { return m_Columns.size(); }
    const CQScoreColumn& GetColumn(unsigned int pos) const {
        return *m_Columns[pos]; }

    double GetLogLike(void) const { return GetLogLike(m_FS->GetParams()); }

    //for debugging
    void PrintObservedSpectrum(bool fold) const;

private:
    class CParser : public xmlpp::SaxParser {
    public:
        CParser(CQScores &scores) :
            m_QScores(scores), m_CacheAssumesAnc(false) {}
    protected:
        virtual void on_start_element(const Glib::ustring &name,
                                      const AttributeList &attList);
        virtual void on_end_element(const Glib::ustring &name);
        virtual void on_characters(const Glib::ustring &text);
    private:
        string m_Inside;
        CQScores &m_QScores; //reference to friend (ug, I know)

        int m_CtgPos;
        string m_PrObservedCond;
        bool   m_CacheAssumesAnc; //how PrObservedCond was calculated
        CQScoreColumn::TScores m_Scores;
        CQScoreColumn::TErrors m_Errors;
    };
    friend class CParser;

    class CFSMaximizer : CGSLSimplex {
    public:
        CFSMaximizer(CQScores &);
        virtual double F(const gsl_vector* x) const {
            CFreqSpectrum::TParams p;
            x_SetParams(x, p);
            return -m_QScores.GetLogLike(p);
        }

    private:
        virtual unsigned int x_GetNumParams(void) const {
            return m_QScores.m_FS->GetNumParams();
        }
        virtual void x_InitializeParams(gsl_vector* x) const {
            x_SetVector(m_QScores.m_FS->GetParams(), x);
        }
        virtual void x_StoreParams(const gsl_vector* x) {
             x_SetParams(x, m_QScores.m_FS->SetParams());
        }

        static
        void x_SetParams(const gsl_vector *x, CFreqSpectrum::TParams &p) {
            p.resize(x->size);
            for (unsigned int i = 0;  i < x->size;  ++i) {
                p[i].first = gsl_vector_get(x, i);
                p[i].second = true;
            }
        }
        static
        void x_SetVector(const CFreqSpectrum::TParams &p, gsl_vector *x) {
            assert(x->size <= p.size());
            for (unsigned int i = 0;  i < x->size;  ++i) {
                gsl_vector_set(x, i, p[i].first);
            }
        }

        CQScores &m_QScores; //reference to friend (ug, I know)
    };
    friend class CFSMaximizer;

private:
    double GetLogLike(const CFreqSpectrum::TParams &) const;

    typedef vector<CQScoreColumn *> TColumns;
    TColumns m_Columns;
    auto_ptr<CFreqSpectrum> m_FS;
    bool m_KnowAncestral;
};

/*---------------------------------------------------------------------------*/
// PRE : reference to frequency spectrum parameters, position within
// contig (not used; just maintained for continuity in serialization)
// minor & major quality scores, optional strings containing P(m-) and
// P(m+)
// POST: P(m_-) and P(m_+) calculated (precursors to other, more
// interesting probabilities)
CQScoreColumn::CQScoreColumn(int contigPos, const TScores &scores,
                             const string &prObservedCond,
                             bool knowAncestral, const TErrors *errors) :
    m_ContigPos(contigPos),
    m_Scores(scores),
    m_pObservedCond(0, m_Scores[eMajor].size() + m_Scores[eMinor].size() +
                    m_Scores[eOther1].size() + m_Scores[eOther2].size())
{
    if (errors) {
        for (int i = 0;  i < 4;  ++i) {
            m_Errors[i] = (*errors)[i];
        }
    }

    if (!prObservedCond.empty()) {
        istringstream iss(prObservedCond);
        m_pObservedCond.DeSerialize(iss);
    } else {
        //pre-process triallelics!
        if (!knowAncestral) {
            EWhichScores mx = eMajor;
            for (int w = 1;  w < 4;  ++w) {
                if (GetDepth((EWhichScores) w) > GetDepth(mx)) {
                    mx = (EWhichScores) w;
                }
            }
            swap(m_Scores[mx], m_Scores[eMajor]);
            swap(m_Errors[mx], m_Errors[eMajor]);
        }
        for (int w = 2;  w < 4;  ++w) { //i.e.  eOther1 and eOther2
            copy(m_Scores[w].begin(), m_Scores[w].end(),
                 back_inserter(m_Scores[eMinor]));
            copy(m_Errors[w].begin(), m_Errors[w].end(),
                 back_inserter(m_Errors[eMinor]));
            m_Scores[w].clear();
            m_Errors[w].clear();
        }

        x_CalcObservedCondDist();
    }
}

/*---------------------------------------------------------------------------*/
// PRE : xml root element, whether we know the ancestral state
// POST: primary column contents (scores, minordepth, P(m_-), P(m_+))
// serialized to XML
void CQScoreColumn::WriteToXML(xmlpp::Element *root, bool knowAnc) const {
    xmlpp::Element *column = root->add_child("Column");
    column->set_attribute("pos", IntToString(m_ContigPos));
    for (unsigned int which = 0;  which < 4;  ++which) {
        for (unsigned int chrI = 0;  chrI < m_Scores[which].size();  ++chrI) {
            xmlpp::Element *chr = column->add_child("Chr");
            chr->set_attribute("score", IntToString(m_Scores[which][chrI]));
            if (GetErrTruth(chrI, (EWhichScores) which)) {
                chr->set_attribute("error", "");
            }
            switch (which) {
            case eMajor: break;
            case eMinor:  chr->set_attribute("minor", "");  break;
            case eOther1: chr->set_attribute("other", "1"); break;
            case eOther2: chr->set_attribute("other", "2"); break;
            }
        }
    }

    xmlpp::Element *prObservedCond = column->add_child("PrObservedCond");
    {
        ostringstream oss;
        m_pObservedCond.Serialize(oss);
        prObservedCond->add_child_text(oss.str());
        if (knowAnc) {
            prObservedCond->set_attribute("anc","");
        }
    }
}

/*---------------------------------------------------------------------------*/
// PRE : boolean state vector
// POST: state "incremented"; whether or not this incrementing "rolled
// over" the counter.  Note this does *NOT* increment in the
// traditional sense as it maintains a constant number of ON bits --
// see code for details
bool CQScoreColumn::x_IncrementState (TState &state) {

    //look for rightmost 10 (if doesn't exist, then we've "rolled over")
    bool found = false;
    int i = state.size() - 1;
    while (i >= 1  &&  !found) {
        if (!state[i]  &&  state[i-1]) {
            found = true;
        } else {
            --i;
        }
    }
    if (!found) { //don't bother incrementing counter since we rolled over
        return false;
    }
    
    //change to 01
    state[i-1] = false;
    state[i] = true;

    //shuffle bits down to pack tightly: 01+0*$
    int j = state.size() - 1;
    while (i < j) {
        if (state[i]) {
            ++i;
        } else if (!state[j]) {
            --j;
        } else {
            state[i] = true;
            state[j] = false;
            ++i;
            --j;
        }
    }
    return true;
}

/*---------------------------------------------------------------------------*/
// PRE : range of depths within column to examine (presumably either
// 0..m_0 or m_0..n)
// POST: probability distribution for errors, given the observed quality scores
void CQScoreColumn::x_CalcErrorDist (EWhichScores which, CProbDist &errorPrbs)
    const {
    //when probabilities reach the "small enough" level, we stop calculating (helps with those high-sample depth guys)
    const double kSmallEnough = 1e-10;
    int depth = GetDepth(which);
    double maxProb = 0;
        
    // loop over all possible values for the # of false alleles
    for (int numFalse = 0;  numFalse <= depth;  ++numFalse) {
        TState errSet(depth);
        fill(errSet.begin(), errSet.begin()+numFalse, true);
        fill(errSet.begin()+numFalse, errSet.end(), false);

        errorPrbs.SetProb(numFalse) = 0;

        if (numFalse == 0  ||
            errorPrbs.GetProb(numFalse-1) > kSmallEnough ||
            maxProb == 0) {//only count if we're going to get a non-negligible probability
            if (gsl_sf_choose(depth, numFalse) > 1e6) {
                cerr << "Depth/numFalse too big at #" << m_ContigPos
                     << "; last prob: " << errorPrbs.GetProb(numFalse-1)
                     << endl;
                continue;
            }

            do {
                double prb = 1;
                
                for (int i = 0;  i < depth;  ++i) {
                    prb *= errSet[i] ?
                        GetErrProb(i, which) : 1 - GetErrProb(i, which);
                }
                
                errorPrbs.SetProb(numFalse) += prb;
            } while (x_IncrementState(errSet));

            maxProb = max(maxProb, errorPrbs.GetProb(numFalse));
        } //end of bothering to count
    }
    
}

/*---------------------------------------------------------------------------*/
// PRE : m_Scores[eMinor]/m_Scores[eMajor] set
// POST: probability of m0 | m stored for all m
void CQScoreColumn::x_CalcObservedCondDist(void) {
    CProbDist pMinus(0, GetDepth(eMinor));
    CProbDist pPlus (0, GetDepth(eMajor));
    x_CalcErrorDist(eMinor, pMinus);
    x_CalcErrorDist(eMajor, pPlus);

    for (unsigned int m = 0;  m <= GetDepth();  ++m) {
        int trueMinor = m;
        int trueMajor = GetDepth() - trueMinor;
        int dm = GetDepth(eMinor) - m; //aka. m0 - m

        double pr = 0;

        if (dm >= 0) {
            int stop = min(trueMinor, trueMajor - dm);
            for (int i = 0;  i <= stop;  ++i) {
                pr += pMinus.GetProb(i+dm) * pPlus.GetProb(i);
            }
        } else {
            int stop = min(trueMinor + dm, trueMajor);
            for (int i = 0;  i <= stop;  ++i) {
                pr += pMinus.GetProb(i) * pPlus.GetProb(i-dm);
            }
        }

        m_pObservedCond.SetProb(m) = pr / gsl_sf_choose(GetDepth(), m);
    }
}

/*---------------------------------------------------------------------------*/
// PRE : m_pObservedCond calculated; actual number of minor alleles (m)
// POST: probability of m0 | m
inline double CQScoreColumn::GetPrObservedCond(int m) const {
    return m_pObservedCond.GetProb(m);
}

/*---------------------------------------------------------------------------*/
// PRE : observed # of minor alleles, params for frequency spectrum
// POST: probability of m0 | frequency spectrum
double CQScoreColumn::GetPrObserved(const CFreqSpectrum &fs) const {
    double pr = 0;

    //Note: loop *intentionally* runs to Depth -- for symmetry which
    //eases handling of folded spectrum
    for (unsigned int m = 0;  m <= GetDepth();  ++m) {
        pr += GetPrObservedCond(m) * GetPrPrior(m, fs);
    }

    return pr;
}

/*---------------------------------------------------------------------------*/
// SAX event-based parser
// PRE :
// POST:
void CQScores::CParser::on_start_element(const Glib::ustring &name,
                                         const AttributeList &attList) {
    if (name == "FSParams") {
        for (AttributeList::const_iterator attI = attList.begin();
             attI != attList.end();  ++attI) {
            if (attI->name == "dist") {
                if (attI->value == "Neutral") {
                    m_QScores.m_FS.reset(new CFSNeutral);
                } else if (attI->value == "Growth") {
                    m_QScores.m_FS.reset(new CFSGrowth);
                }
            } else if (attI->name == "params") {
                m_QScores.m_FS->SetParams(attI->value);
            }
        }
    } else if (name == "Column") {
        m_CacheAssumesAnc = false;
        m_PrObservedCond.clear();
        for (int i = 0;  i < 4;  ++i) {
            m_Scores[i].clear();
            m_Errors[i].clear();
        }
        m_CtgPos = atoi(attList.front().value.c_str());
    } else if (name == "Chr") {
        int score;
        CQScoreColumn::EWhichScores which = CQScoreColumn::eMajor;
        bool error = false;
        for (AttributeList::const_iterator attI = attList.begin();
             attI != attList.end();  ++attI) {
            if (attI->name == "score") {
                score = atoi(attI->value.c_str());
            } else if (attI->name == "minor") {
                which = CQScoreColumn::eMinor;
            } else if (attI->name == "other") {
                which = (attI->value == "1") ?
                    CQScoreColumn::eOther1 : CQScoreColumn::eOther2;
            } else if (attI->name == "error") {
                error = true; 
            }
        }
        m_Scores[which].push_back(score);
        m_Errors[which].push_back(error);
    } else if (name == "PrObservedCond") {
        for (AttributeList::const_iterator attI = attList.begin();
             attI != attList.end();  ++attI) {
            if (attI->name == "anc") {
                m_CacheAssumesAnc = true;
            }
        }
    }
    m_Inside = name;
}
void CQScores::CParser::on_end_element(const Glib::ustring &name) {
    if (name == "Column") {
        if (m_CacheAssumesAnc != m_QScores.GetKnowAncestral()) {
            m_PrObservedCond = "";
        }

        m_QScores.m_Columns.push_back
            (new CQScoreColumn(m_CtgPos, m_Scores, m_PrObservedCond,
                               m_QScores.GetKnowAncestral(), &m_Errors));
    }
    m_Inside = "";
}
void CQScores::CParser::on_characters(const Glib::ustring &text) {
    if (m_Inside == "PrObservedCond") {
        m_PrObservedCond = text;
    }
}

/*---------------------------------------------------------------------------*/
// PRE : open input stream with serialized QScore data (format TBA)
// POST: data parsed & loaded into memory
void CQScores::LoadFromStream(istream &in) {
    if (!in) {
        throw runtime_error("CQScores::LoadFromStream: stream not Good.");
    }

    CParser parser(*this);
#ifdef _DEBUG
    parser.set_validate();
#endif
    parser.parse_stream(in);
}

/*---------------------------------------------------------------------------*/
// PRE : open output stream
// POST: data serialized & sent to stream
void CQScores::WriteToStream(ostream &out) {
    if (!out) {
        throw runtime_error("CQScores::WriteToStream: stream not Good.");
    }
    
    xmlpp::Document doc;
    doc.set_internal_subset("Contig", "", "qual.dtd");
    xmlpp::Element *root = doc.create_root_node("Contig");

    for (TColumns::const_iterator colI = m_Columns.begin();
         colI != m_Columns.end();  ++colI) {
        (*colI)->WriteToXML(root, m_KnowAncestral);
    }

    xmlpp::Element *params = root->add_child("FSParams");
    params->set_attribute("dist", m_FS->GetName());
    string paramStr;
    for (unsigned int paramI = 0;  paramI < m_FS->GetNumParams();  ++paramI) {
        paramStr += DblToString(m_FS->GetParam(paramI)) + " ";
    }
    params->set_attribute("params", paramStr);

    out << doc.write_to_string_formatted();
}

/*---------------------------------------------------------------------------*/
// PRE : qscore data loaded; frequency spectrum params
// POST: log likelihood of the data given these parameters
double CQScores::GetLogLike(const CFreqSpectrum::TParams &p) const {
    auto_ptr<CFreqSpectrum> fs(m_FS->Clone());
    fs->SetParams() = p;
    if (fs->ParamsOutsideBounds()) {
        return log(0.);
    }

    double llik = 0;
    for (unsigned int colI = 0;  colI < GetNumColumns();  ++colI) {
        double pr = GetColumn(colI).GetPrObserved(*fs);
        llik += log(pr);
    }

    assert(!isnan(llik));
    return llik;
}

/*---------------------------------------------------------------------------*/
// PRE : qScores object
// POST: frequency spectrum parameters set to maximize log likelihood
CQScores::CFSMaximizer::CFSMaximizer(CQScores &qScores) :
    m_QScores(qScores)
{
    if (m_QScores.m_FS->GetNumParams() == 0) {
        return;
    }

    Optimize(.01, .00001, 200);
}

/*---------------------------------------------------------------------------*/
// PRE : qscore data loaded, start/stop ranges for params, increment
// POST: parameters set quasi-optimally (?)
void CQScores::OptimizeParamsGrid(const CFreqSpectrum::TParams &start,
                                  const CFreqSpectrum::TParams &stop,
                                  const CFreqSpectrum::TParams &inc) {
    CFreqSpectrum::TParams maxParams, p;


    maxParams = start;
    double maxLik = GetLogLike(start);

    p = start;
    do {
        //want to maximize log likelihood:
        double lik = GetLogLike(p);
        cout << p << "\t" << lik << endl;
        if (lik > maxLik) {
            maxLik = lik;
            maxParams = p;
        }

        for (unsigned int paramI = 0;  paramI < p.size();  ++paramI) {
            if (p[paramI] < stop[paramI]) {
                p[paramI].first += inc[paramI].first;
                break;
            } else {
                p[paramI] = start[paramI];
            }
        }
    } while (p != start);

    m_FS->SetParams() = maxParams;
}


/*---------------------------------------------------------------------------*/
// PRE : qscore data loaded; whether or not to fold observed spectrum
// POST: observed fs sent to stdout (mostly for debugging)
void CQScores::PrintObservedSpectrum(bool fold) const {
    vector< vector<double> > fs;
    for (unsigned int colI = 0;  colI < GetNumColumns();  ++colI) {
        unsigned int depth = GetColumn(colI).GetDepth();
        if (depth+1 > fs.size()) {
            fs.resize(depth+1);
        }
        fs[depth].resize(depth+1, 0);
        if (GetColumn(colI).GetDepth(CQScoreColumn::eOther1) == 0  &&
            GetColumn(colI).GetDepth(CQScoreColumn::eOther2) == 0) {
            ++fs[depth][GetColumn(colI).GetDepth(CQScoreColumn::eMinor)];
        }
    }

    for (unsigned int depth = 1;  depth < fs.size();  ++depth) {
        double tot = 0;
        for (unsigned int i = 0;  i < fs[depth].size();  ++i) {
            tot += fs[depth][i];
        }
        if (tot > 0) {
            cout << "Depth: " << depth << endl;
            cout << "  " << tot << ":";

            if (!fold) {
                for (unsigned int i = 0;  i < fs[depth].size();  ++i) {
                    cout << " " << fs[depth][i]/tot;
                }
            } else {
                for (unsigned int i = 0;  i < (fs[depth].size()+1)/2;  ++i) {
                    if (depth - i == i) {
                        cout << " " << fs[depth][i]/tot;
                    } else {
                        cout << " " << (fs[depth][i] + fs[depth][depth-i])/tot;
                    }
                }
            }
            cout << endl;
        }
    }
}

/*---------------------------------------------------------------------------*/

int Main(TOpts &opts) {
    CQScores qscores(new CFSNeutral);
    qscores.SetKnowAncestral() = (opts.find('a') != opts.end());

    ifstream infile(opts['i']);
    qscores.LoadFromStream(string(opts['i']) == "-" ? cin : infile);

    if (opts.find('f') != opts.end()) {
        if (string(opts['f']) == "neutral") {
            qscores.SetFreqSpectrum().reset(new CFSNeutral);
        } else if (string(opts['f']) == "growth") {
            qscores.SetFreqSpectrum().reset(new CFSGrowth);
        } else {
            throw runtime_error(string("Invalid frequency spectrum '") +
                                opts['f'] + "'.");
        }
    }
    qscores.SetFreqSpectrum()->SetFolded(opts.find('a') == opts.end());

    if (opts.find('p') != opts.end()) {
        qscores.OptimizeParamsSimplex();
    }

    if (opts.find('g') != opts.end()) { //explore likelihood [g]rid
        CFreqSpectrum::TParams start, stop, inc;
        start.push_back(CFreqSpectrum::TParams::value_type(0.010,true));
        stop.push_back(CFreqSpectrum::TParams::value_type (0.020,true));
        inc.push_back(CFreqSpectrum::TParams::value_type  (0.001,true));

        start.push_back(CFreqSpectrum::TParams::value_type(25,true));
        stop.push_back(CFreqSpectrum::TParams::value_type (75,true));
        inc.push_back(CFreqSpectrum::TParams::value_type  (5,true));

        qscores.OptimizeParamsGrid(start, stop, inc);
    }

#ifdef _DEBUG
    qscores.PrintObservedSpectrum(qscores.SetFreqSpectrum()->IsFolded());
#endif
    cout.precision(8);
    cout << "LogLikelihood: " << qscores.GetLogLike() << " / "
         << qscores.GetFreqSpectrum().GetName() << " w/ "
         << qscores.GetFreqSpectrum().GetParams() << endl;

    if (opts.find('o') != opts.end()) {
        ofstream outfile(opts['o']);
        qscores.WriteToStream(outfile);
    }

    if (opts.find('1') != opts.end()) {
        cout.precision(2);
        for (unsigned int colI = 0;  colI < qscores.GetNumColumns();  ++colI) {
            const CQScoreColumn &col = qscores.GetColumn(colI);
            if (col.GetDepth() > 1) {
                if (col.GetDepth(CQScoreColumn::eMinor) == 1) {
                    cout << col.GetErrProb(0, CQScoreColumn::eMinor) << "\t"
                         << col.GetPrPosterior(0, qscores.GetFreqSpectrum()) <<"\t"
                         << col.GetErrTruth(0, CQScoreColumn::eMinor)
                         << endl;
                } else if (qscores.GetFreqSpectrum().IsFolded()  &&
                           col.GetDepth(CQScoreColumn::eMajor) == 1) {
                    cout << col.GetErrProb(0, CQScoreColumn::eMajor) << "\t"
                         << col.GetPrPosterior(col.GetDepth(),
                                               qscores.GetFreqSpectrum()) << "\t"
                         << col.GetErrTruth(0, CQScoreColumn::eMajor)
                         << endl;
                }
            }
        }
    }

    if (opts.find('2') != opts.end()) {
        cout.precision(2);
        for (unsigned int colI = 0;  colI < qscores.GetNumColumns();  ++colI) {
            const CQScoreColumn &col = qscores.GetColumn(colI);
            if (col.GetDepth(CQScoreColumn::eMinor) == 2) {
                double q1 = col.GetErrProb(0, CQScoreColumn::eMinor);
                double q2 = col.GetErrProb(1, CQScoreColumn::eMinor);
                cout << q1*q2 << "\t"
                     << col.GetPrPosterior(0, qscores.GetFreqSpectrum())
                     << endl;
            }
        }
    }

    if (opts.find('3') != opts.end()) {
        cout.precision(2);
        for (unsigned int colI = 0;  colI < qscores.GetNumColumns();  ++colI) {
            const CQScoreColumn &col = qscores.GetColumn(colI);
            if (col.GetDepth(CQScoreColumn::eMinor) == 3) {
                double q1 = col.GetErrProb(0, CQScoreColumn::eMinor);
                double q2 = col.GetErrProb(1, CQScoreColumn::eMinor);
                double q3 = col.GetErrProb(2, CQScoreColumn::eMinor);
                cout << q1*q2*q3 << "\t"
                     << col.GetPrPosterior(0, qscores.GetFreqSpectrum())
                     << endl;
            }
        }
    }

    return 0;
}

/*---------------------------------------------------------------------------*/
int main (int argc, char** argv) {
    try {
        TOpts opts;
        x_ProcessOptions
            (argc, argv, "i:o:f:gap123", "i", opts,
             " -i <input.xml> [-o <output.xml>] [-f <neutral|growth>] [-p123] [-c <#points in ci>] \n"
             "Where\t-i --> input XML filename\n"
             "\t-f --> (optional) frequency spectrum distribution\n"
             "\t-o --> (optional) output XML filename\n"
             "\t-g --> grid likelihood landscape\n"            
             "\t-a --> flag to use [a]ll of spectrum w/o folding\n"
             "\t-p --> flag to optimize parameters\n"
             "\t-1 --> flag to show pre/post error probs for singletons\n"
             "\t-2 --> flag to show pre/post error probs for doubltons\n"
             "\t-3 --> flag to show pre/post error probs for tripletons\n");
        
        return Main(opts);
    } catch (exception &e) {
        cerr << "Exception:\n\t" << e.what() << endl;
        return 1;
    }
}
